import pytest

from commonnexus.tokenizer import *


@pytest.mark.parametrize(
    'text,ttype',
    [
        ("abcd", TokenType.WORD),
        ("'a b'", TokenType.QWORD),
        (" \t\n", TokenType.WHITESPACE),
        ('+', TokenType.PUNCTUATION),
        ("[a 'comment+]", TokenType.COMMENT),
    ]
)
def test_iter_tokens(text, ttype):
    tokens = list(iter_tokens(iter(text)))
    assert len(tokens) == 1 and tokens[0].type == ttype


@pytest.mark.parametrize(
    'text,words',
    [
        ("Bembidion\nB._zephyrum\n'John''s sparrow (eastern) '\n",
         ['Bembidion', 'B._zephyrum', "John's sparrow (eastern) "]),
        (r"[\\i]Bembidion_velox[\\p]_Linnaeus", ['Bembidion_velox_Linnaeus']),
        ("'B. zephyrum' ", [Word('B._zephyrum')]),
        ("aword 'the ''word'", ["aword", "the 'word"]),
    ]
)
def test_iter_words(text, words):
    assert [w for w in iter_words_and_punctuation(iter_tokens(iter(text)))] == words


def test_iter_words_2():
    assert [w for w in iter_words_and_punctuation(iter_tokens(iter('abc-def')), allow_punctuation_in_word='-')] == ['abc-def']


@pytest.mark.parametrize(
    'start,string,length,allow_single_word',
    [
        ('"', 'a b c"', 3, False),
        ('word', 'a b c"', 1, True),
        (None, '"a b c"', 3, False),
    ]
)
def test_iter_delimited(start, string, length, allow_single_word):
    res = list(iter_delimited(
        start,
        iter_words_and_punctuation(iter_tokens(iter(string))),
        allow_single_word=allow_single_word))
    assert len(res) == length



import pytest

from commonnexus.tools.combine import combine


@pytest.fixture
def nex1(nexus):
    return nexus(DATA="""
            Dimensions ntax=2 nchar=1;
            Format datatype=standard symbols="12" gap=-;
            Matrix
            Harry              1
            Simon              2
            ;""")


@pytest.fixture
def nex2(nexus):
    return nexus(DATA="""
            Dimensions ntax=2 nchar=1;
            Format datatype=standard symbols="34" gap=-;
            Matrix
            Harry              3
            Simon              4
            ;""")


@pytest.fixture
def nex3(nexus):
    return nexus(DATA="""
            Dimensions ntax=3 nchar=1;
            Format datatype=standard symbols="345" gap=-;
            Matrix
            Betty              3
            Boris              4
            Simon              5
            ;""")


def test_combine_simple(nex1, nex2):
    matrix = combine(nex1, nex2).characters.get_matrix()
    assert matrix['Harry'] == {'1.1': '1', '2.1': '3'}
    assert matrix['Simon'] == {'1.1': '2', '2.1': '4'}


def test_combine_missing(nex1, nex3):
    matrix = combine(nex1, nex3).characters.get_matrix()
    assert matrix['Harry'] == {'1.1': '1', '2.1': None}
    assert matrix['Simon'] == {'1.1': '2', '2.1': '5'}
    assert matrix['Betty'] == {'1.1': None, '2.1': '3'}
    assert matrix['Boris'] == {'1.1': None, '2.1': '4'}


def test_combine_iterated(nex1):
    res = combine(combine(nex1))
    assert len(res.taxa) == 2  # Taxa are merged.
    assert '1.1.1' in res.characters.get_matrix()['Simon']  # Charlabels are prefixed.


def test_combine_with_character_labels(nexus):
    n1 = nexus(
        TAXA="DIMENSIONS NTAX=3; TAXLABELS Tax1 Tax2 Tax3;",
        DATA="""
            DIMENSIONS NCHAR=3;
            FORMAT DATATYPE=STANDARD MISSING=0 GAP=-  SYMBOLS="123";
            CHARSTATELABELS
                1 char1,
                2 char2,
                3 char3
        ;
        MATRIX
        1         123
        2         123
        3         123
        ;
        """)
    n2 = nexus(DATA="""
            DIMENSIONS NTAX=3 NCHAR=3;
            FORMAT DATATYPE=STANDARD MISSING=0 GAP=-  SYMBOLS="456";
            CHARSTATELABELS
                1 char1,
                2 char2,
                3 char3
        ;
        MATRIX
        Tax1         456
        Tax2         456
        Tax3         456
        ;
        """)
    newnex = combine(n1, n2)
    assert len(newnex.taxa) == 3, "Taxa not merged using taxlabels"
    row = list(newnex.characters.get_matrix().values())[0]
    assert len(row) == 6, "Characters not aggregated"
    assert '1.char1' in row and ('2.char1' in row), "charlabels not prefixed"


def test_combine_trees(nexus):
    nex1 = nexus(TREES="translate 1 a, 2 b, 3 c; tree 1 = (1,2,3);")
    nex2 = nexus(TREES="""
            tree 2 = (b,a[comment],c);
            tree 3 = (b,c,a);""")

    newnex = combine(nex1, nex2)
    assert len(newnex.TREES.trees) == 3, "not all trees aggregated"
    assert newnex.taxa == ['a', 'b', 'c'], "trees not properly translated"
    assert newnex.TREES.trees[0].name == '1.1', "tree name not prefixed"
    assert '[comment]' in newnex.TREES.trees[1].newick.newick, "comment in newick lost"



import pytest

from commonnexus.blocks.notes import Notes


@pytest.mark.parametrize(
    'spec,expect',
    [
        (dict(
            TAXA="DIMENSIONS NTAX=5; TAXLABELS t1 't2 x' t3 t4 t5;",
            CHARACTERS="DIMENSIONS NCHAR=3; CHARSTATELABELS 1 c1/c1.0 c1.1, 2 c2, 3 c3;",
            NOTES="TEXT SOURCE=INLINE TAXON=(1-3) CHARACTER=1 STATE=1 TEXT= 'these taxa from north';"),
         lambda n: n.TEXT.taxons == ['t1', "t2 x", 't3'] and n.TEXT.characters == ['c1'] and n.TEXT.states == {'c1': ['c1.1']}),
        (dict(
            TREES="TREE 'the tree' = (,):2;",
            CHARACTERS="""
DIMENSIONS NEWTAXA NTAX=5 NCHAR=3;
TAXLABELS t1 't2 x' t3 t4 t5;
CHARSTATELABELS 1 c1/c1.0 c1.1, 2 c2, 3 c3;""",
            NOTES="TEXT SOURCE=INLINE TAXON=(1 3) TREE=1 TEXT= 'these taxa from the far north';"),
         lambda n: n.TEXT.taxons == ['t1', 't3'] and n.TEXT.trees == ['the tree']),
        (dict(
            CHARACTERS="DIMENSIONS NEWTAXA NTAX=5 NCHAR=3;",
            NOTES="TEXT SOURCE=INLINE TAXON=(1 3) TEXT= 'these taxa from the far north';"),
         lambda n: n.TEXT.taxons == ['1', '3']),
    ]
)
def test_Notes_(nexus, spec, expect):
    nex = nexus(**spec)
    assert expect(nex.NOTES)


def test_Notes_from_data(nexus):
    nex = nexus(TAXA="DIMENSIONS NTAX=5; TAXLABELS A B C D E;")
    nex.append_block(Notes.from_data([dict(taxons=['A', 'D'], text='Some stuff')], nexus=nex))
    assert "TEXT TAXON=(1 4)" in str(nex) and "TEXT='Some stuff'" in str(nex)

    nex = nexus(DATA="DIMENSIONS NCHAR=3; CHARLABELS X Y Z; MATRIX t1 101 t2 010 t3 001;")
    nex.append_block(Notes.from_data([dict(characters=['Y', 'Z'], text='Some stuff')], nexus=nex))
    assert "TEXT CHARACTER=(2 3)" in str(nex) and "TEXT='Some stuff'" in str(nex)

    nex = nexus(TREES="TREE x = (a,b); TREE y = (a,b); TREE z = (a,b);")
    nex.append_block(Notes.from_data([dict(trees=['x', 'z'], text='Some stuff')], nexus=nex))
    assert "TEXT TREE=(1 3)" in str(nex) and "TEXT='Some stuff'" in str(nex)



import pytest

from commonnexus import Nexus
from commonnexus.blocks.trees import Trees


def test_Trees(nexus):
    nex = nexus(
        TAXA="TAXLABELS Scarabaeus Drosophila Aranaeus;",
        TREES="""\
    TRANSLATE beetle Scarabaeus, fly Drosophila, spider Aranaeus;
    TREE tree1 = [&R] ((beetle,fly),spider);
    TREE tr[comment]ee-2 = ((1,2),3);
    TREE tree3 = ((Scarabaeus,Drosophila),Aranaeus);"""
    )
    assert nex.TREES.TRANSLATE.mapping['spider'] == 'Aranaeus'
    assert nex.TREES.TREE.newick.newick == '((beetle,fly),spider)'
    assert nex.TREES.TREE.rooted
    node = nex.TREES.translate(nex.TREES.TREE)
    assert node.newick == '((Scarabaeus,Drosophila),Aranaeus)'
    tree = nex.TREES.trees[1]
    assert tree.name == 'tree-2'
    assert nex.TREES.translate(tree).newick == '((Scarabaeus,Drosophila),Aranaeus)'


def test_Trees_with_leading_comment(nexus):
    nex = nexus(
        TAXA="TAXLABELS Scarabaeus Drosophila Aranaeus;",
        TREES="TREE tree1 = [&R] [&clockrate=9.118820e-04] ((beetle,fly),spider);")
    assert nex.TREES.TREE.rooted
    assert '&clockrate=9.118820e-04' in nex.comments


def test_Tree_translate_with_taxa_from_linked_block(nexus):
    nex = nexus(
        TAXA="TITLE x; TAXLABELS A B C;",
        TREES="LINK TAXA=x; TREE t = (1,2)3;"
    )
    assert nex.TREES.translate(nex.TREES.TREE).newick == '(A,B)C'


def test_Tree_translate_with_numeric_labels():
    nex = Nexus("""#NEXUS
    BEGIN TAXA;
        TAXLABELS Scarabaeus Drosophila Aranaeus;
    END;
    BEGIN TREES;
        TRANSLATE 0 Scarabaeus, 1 Drosophila, 2 Aranaeus;
        TREE tree = ((0,1),2);
    END;""")
    tree = nex.TREES.translate(nex.TREES.TREE)
    assert {n.name for n in tree.walk() if n.name} == {'Scarabaeus', 'Drosophila', 'Aranaeus'}


def test_Trees_complex_newick():
    s = """#nexus
BEGIN TAXA;
    TAXLABELS A B 'And C';
END;
BEGIN Trees;
    TREE tree = (A[commentA]:1.1,B[commentB]:2.2,X-1)'And C'[comment C]:3.3;
END;
"""
    nex = Nexus(s)
    nodes = {n.name: n for n in nex.TREES.TREE.newick.walk()}
    assert nodes['A'].comment == 'commentA' and nodes['A'].length == pytest.approx(1.1)
    assert nodes['B'].comment == 'commentB' and nodes['B'].length == pytest.approx(2.2)
    assert nodes["'And C'"].comment == 'comment C' and nodes["'And C'"].length == pytest.approx(3.3)

    nex = Nexus(s, validate_newick=True)
    nodes = {n.name: n for n in nex.TREES.TREE.newick.walk()}
    assert nodes['A'].comment == 'commentA' and nodes['A'].length == pytest.approx(1.1)
    assert nodes['B'].comment == 'commentB' and nodes['B'].length == pytest.approx(2.2)
    assert nodes["'And C'"].comment == 'comment C' and nodes["'And C'"].length == pytest.approx(3.3)


def test_Trees_from_data(nexus):
    nex = nexus(TAXA="TAXLABELS A B C;")
    nex.append_block(Trees.from_data(('the tree', '(X,Y)Z', False), A='X', B='Y', C='Z'))
    assert str(nex) == """#NEXUS
BEGIN TAXA;
TAXLABELS A B C;
END;
BEGIN TREES;
TRANSLATE A X,
B Y,
C Z;
TREE 'the tree' = [&U] (A,B)C;
END;"""


@pytest.mark.parametrize(
    'trees',
    [
        "INVALID;",
        "TRANSLATE; TRANSLATE;"
    ]
)
def test_Trees_validate(nexus, trees):
    with pytest.raises(ValueError):
        _ = nexus(TREES=trees).validate()


def test_Tree_newick_string(nexus):
    nex = nexus(TREES="TREE x = (a, b, c)d;")
    trees = nex.TREES
    assert trees.TREE.newick_string == '(a, b, c)d;'
    assert {n.name for n in trees.TREE.newick.walk()} == set('abcd')


def test_Trees_with_comment():
    nex = Nexus('#NEXUS\n')
    nex.append_block(Trees.from_data(comment='the comment'))
    assert '[the comment]' in str(nex)


def test_BEAST_like_trees():
    nex = Nexus('#NEXUS\n')
    nex.append_block(Trees.from_data(
        ('STATE_1', '(X,Y)Z', False), lowercase_command=True))
    # tracerer counts trees using this pattern:
    # pattern = "(^tree STATE_)|(\tTREE \\* UNTITLED = \\[&R\\] \\()")
    assert any(line.startswith('tree STATE_1') for line in str(nex).split('\n'))



"""
\3 -> every third for NUCLEOTIDE
tests/data/chars/primates.chars.subsets-all.nexus
begin sets;
>-------charset coding = 2-457 660-896;
>-------charset noncoding = 1 458-659 897-898;
>-------charset 1stpos = 2-457\3 660-896\3;
>-------charset 2ndpos = 3-457\3 661-896\3;
>-------charset 3rdpos = 4-457\3 662-.\3;
end;

tests/data/chars/orti.nex
BEGIN SETS;
   TaxSet EPAC = 1-14;
   TaxSet WPAC = 15-23;
END;

tests/data/chars/COII_Apes.nex
BEGIN SETS;
   TaxSet Human = 1-6;
   TaxSet Chimp = 7-11;
   TaxSet PygmyChimp = 12-15;
   TaxSet Gorilla = 16-21;
   TaxSet Orangutan = 22-24;
END;
"""


import logging

import pytest

from commonnexus import Nexus, Block
from commonnexus.tokenizer import TokenType


@pytest.mark.parametrize(
    'nex,expect',
    [
        (
                """#NEXUS
    BEGIN TREES;
    TREE best = (fish , (frog,
    (snake, mouse))) ;
    END;""",
                lambda n: sum(sum(1 for t in c if t.type != TokenType.WHITESPACE) for c in n) == 22),
        (
                '#NEXUS begin trees; tree ; end;',
                lambda n: len(n.blocks['TREES']) == 1),
        (
                '#NEXUS begin distances; matrix t1 0 t2 1 0 t3 2 1 0; end;',
                lambda n: len(n.taxa) == 3),
        (
                "#NEXUS BEGIN AssuMP[co[mm]ent]TiONS; ENDblock;",
                lambda n: 'ASSUMPTIONS' in n.blocks and n.comments[0] == 'co[mm]ent'),
        (
            "#NEXUS begin block; cmd [comment]; end;",
            lambda n: n.BLOCK.CMD.comments == ['comment'],
        ),
        # Commands cannot contain semicolons, except as terminators, unless the semicolons are
        # contained within a comment or within a quoted token consisting of more than one text
        # character.
        (
                "#NEXUS begin block; cmd [;] stuff; end;",
                lambda n: next(n.blocks['BLOCK'][0][1].iter_payload_tokens(TokenType.WORD)).text == 'stuff'),
        (
                "#NEXUS begin block; cmd 'a;b' stuff; end;",
                lambda n: next(n.blocks['BLOCK'][0][1].iter_payload_tokens(TokenType.WORD)).text == 'stuff'),
        (
                "#NEXUS begin block; cmd '[ab]' stuff; end;",
                lambda n: not n.comments),
        # Single quotes in quoted words must be doubled:
        (
                "#NEXUS begin block; cmd 'a''b' stuff; end;",
                lambda n: "a'b" in [t.text for t in n.blocks['BLOCK'][0][1].iter_payload_tokens()]),
        (
                "#NEXUS begin block; cmd first\tline \t\n second line; end;",
                lambda n: n.BLOCK.commands['CMD'][0].lines == ['first\tline', 'second line']),
        # Leave newick payload untouched.
        (
                '#NEXUS begin trees; stuff (a[&comment]:1,b:2)c:3; end trees;',
                lambda n: str(n.TREES.commands['STUFF'][0]) == '(a[&comment]:1,b:2)c:3'),
        (
                '#NEXUS begin trees; tree t = (a[&comment]:1,b:2)c:3; end trees;',
                lambda n: [nn.name for nn in n.TREES.TREE.newick.walk() if nn.is_leaf] == ['a', 'b']),
        # Test on "real" Nexus files:
        (
                'ape_random.trees',
                lambda n: n.comments[0] == 'R-package APE, Mon Apr  4 13:30:05 2011' and 'TAXA' in n.blocks),
        (
                'quoted_tree_name.nex',
                lambda n: n.TREES.TREE.name.startswith('Transformed') and n.TREES.TREE.newick),
        (
                'regression/detranslate-beast-2.trees',
                lambda n: str(n.TREES.TREE).startswith('TREE1 = [&R] ((((4[&length')),
        (
                'regression/example.trees',
                lambda n: str(n.blocks['TREES'][0].commands['TREE'][2]).endswith('David:0.3086497606)')
        )
    ]
)
def test_Nexus(nex, expect, fixture_dir):
    if nex.startswith('#'):
        n = Nexus(nex)
    else:
        p = fixture_dir / nex
        n = Nexus.from_file(p)
        nex = p.read_text(encoding='utf8')
    assert expect(n)
    assert str(n).strip() == nex.strip(), 'Round-tripping failed!'


@pytest.mark.parametrize(
    'spec,labels,resolved',
    [
        ('2', 't1 t2', ['t2']),
        ('2 - 4', 't1 t2 t3 t4', ['t2', 't3', 't4']),
        ('1 3 - 4', 't1 t2 t3 t4', ['t1', 't3', 't4']),
        ('2 - .', 't1 t2 t3 t4', ['t2', 't3', 't4']),
    ]
)
def test_Nexus_resolve_set_spec(nexus, spec, labels, resolved):
    nex = nexus(TAXA="DIMENSIONS NTAX={}; TAXLABELS {};".format(len(labels.split()), labels))
    assert nex.resolve_set_spec('TAXON', spec.split()) == resolved


def test_Nexus_modification():
    nex = Nexus.from_blocks(Block.from_commands([('CMD', 'stuff')]))
    assert str(nex.BLOCK.CMD) == 'stuff'
    assert nex.BLOCK.OTHER is None
    nex.append_command(nex.BLOCK, 'other')
    assert nex.BLOCK.OTHER
    nex.remove_block(nex.BLOCK)
    assert nex.BLOCK is None


def test_Nexus_replace_block():
    nex = Nexus("""#nexus
    BEGIN TAXA;
        TAXLABELS Scarabaeus Drosophila Aranaeus;
    END;
    BEGIN T[c]rees;
        TRANSLATE beetle Scarabaeus, fly Drosophila, spider Aranaeus;
        TREE tree1 = ((beetle,fly),spider);
        TREE tree2 = ((1,2),3);
        TREE tree3 = ((Scarabaeus,Drosophila),Aranaeus);
    END;""")
    trees = []
    for tree in nex.TREES.trees:
        trees.append(nex.TREES.translate(tree).newick)
    nex.replace_block(
        nex.TREES, [('TREE', 'tree{} = {}'.format(i + 1, n)) for i, n in enumerate(trees)])
    assert str(nex) == """#NEXUS
    BEGIN TAXA;
        TAXLABELS Scarabaeus Drosophila Aranaeus;
    END;
BEGIN TREES;
TREE tree1 = ((Scarabaeus,Drosophila),Aranaeus);
TREE tree2 = ((Scarabaeus,Drosophila),Aranaeus);
TREE tree3 = ((Scarabaeus,Drosophila),Aranaeus);
END;"""


def test_Nexus_validate(caplog):
    nex = Nexus('#NEXUS begin trees; tree t = (a,b)c; translate 1 a, 2 b, 3 c; end;')
    assert not nex.validate(logging.getLogger(__name__))
    assert len(caplog.records) == 1


def test_Config(fixture_dir):
    nex = Nexus.from_file(fixture_dir / 'christophchamp_basic.nex')
    assert nex.TREES.TREE.name == 'basic bush'


def test_encoding_guesser(fixture_dir, tmp_path):
    nex = Nexus.from_file(fixture_dir / 'encoding.nex')
    assert nex.cfg.encoding == 'latin1'
    o = tmp_path / 'test.nex'
    nex.to_file(o)
    assert o.read_text(encoding='latin1') == str(nex)


def test_Booleans_With_Values(fixture_dir):
    nex = Nexus.from_file(fixture_dir / 'christophchamp_dna.nex')
    assert nex.DATA.FORMAT.interleave
    assert len(nex.DATA.get_matrix()['Cow']) == 705


def test_Mesquite_multitaxa(fixture_dir):
    nex = Nexus.from_file(fixture_dir / 'multitaxa_mesquite.nex')
    assert [char.get_matrix() for char in nex.blocks['CHARACTERS']]


def test_comments():
    nex = Nexus('#NEXUS\n[comment]')
    assert nex.comments == ['comment']
    nex.append_block(Block.from_commands([], name='b', comment='block comment'))
    assert nex.comments == ['comment', 'block comment']

    with pytest.raises(ValueError):
        nex = Nexus('#NEXUS\n[comment]]')
        nex.validate()

    with pytest.raises(ValueError):
        nex = Nexus('#NEXUS\nBEGIN b; end;[comment]]')
        nex.validate()



import pytest

from commonnexus import Nexus


@pytest.mark.dendropy
def test_dendropy_suite(dendropyexample):
    n = Nexus(dendropyexample, ignore_unsupported=True)
    assert str(n).strip() == dendropyexample.strip()
    for name, blocks in n.blocks.items():
        for block in blocks:
            _ = block.commands  # We have to access `commands` to actually parse command payloads.
    if n.characters:
        if n.characters.FORMAT and n.characters.FORMAT.datatype == 'CONTINUOUS':
            pass
        else:
            _ = n.characters.get_matrix()
    if n.TREES:
        for tree in n.TREES.trees:
            _ = tree.newick
    assert not n.DISTANCES



from commonnexus import Nexus
from commonnexus.tools.normalise import normalise


def test_normalise(nexus):
    nex = nexus(
        CHARACTERS="DIMENSIONS NCHAR=3; MATRIX 't 1' 100 t2 010 t3 001;",
        DISTANCES="FORMAT NODIAGONAL; MATRIX 't 1' t2  1.0 t3 2.0 3.0;",
        TREES="TRANSLATE a 't 1', b t2, c t3; TREE 1 = (a,b\n,c);")
    res = str(normalise(nex))
    assert res == """#NEXUS
BEGIN TAXA;
DIMENSIONS NTAX=3;
TAXLABELS 't 1' t2 t3;
END;
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
MATRIX 
't 1' 100
t2    010
t3    001
;
END;
BEGIN DISTANCES;
DIMENSIONS NTAX=3;
FORMAT TRIANGLE=BOTH MISSING=?;
MATRIX 
't 1' 0 1.0 2.0
t2    1.0 0 3.0
t3    2.0 3.0 0
;
END;
BEGIN TREES;
TREE 1 = ('t 1',t2,t3);
END;"""
    res = normalise(Nexus(res))
    assert res.characters.get_matrix()
    assert res.DISTANCES.get_matrix()

    res.remove_block(res.TAXA)
    res.remove_block(res.CHARACTERS)
    res = normalise(res)
    assert res.TAXA


def test_normalise_remove_taxon(nexus):
    nex = nexus(
        CHARACTERS="DIMENSIONS NCHAR=3; MATRIX 't 1' 100 t2 010 t3 001;",
        DISTANCES="FORMAT NODIAGONAL; MATRIX 't 1' t2  1.0 t3 2.0 3.0;",
        TREES="TRANSLATE a 't 1', b t2, c t3; TREE 1 = (a,b\n,c);")
    res = str(normalise(nex, remove_taxa={'t2'}))
    assert res == """#NEXUS
BEGIN TAXA;
DIMENSIONS NTAX=2;
TAXLABELS 't 1' t3;
END;
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
MATRIX 
't 1' 100
t3    001
;
END;
BEGIN DISTANCES;
DIMENSIONS NTAX=2;
FORMAT TRIANGLE=BOTH MISSING=?;
MATRIX 
't 1' 0 2.0
t3    2.0 0
;
END;
BEGIN TREES;
TREE 1 = ('t 1',t3);
END;"""


def test_normalise_rename_taxon1(nexus):
    nex = nexus(
        CHARACTERS="DIMENSIONS NCHAR=3; MATRIX 't 1' 100 t2 010 t3 001;",
        DISTANCES="FORMAT NODIAGONAL; MATRIX 't 1' t2  1.0 t3 2.0 3.0;",
        TREES="TRANSLATE a 't 1', b t2, c t3; TREE 1 = (a,b\n,c);")
    res = str(normalise(nex, rename_taxa={'t 1': 't1'}))
    assert res == """#NEXUS
BEGIN TAXA;
DIMENSIONS NTAX=3;
TAXLABELS t1 t2 t3;
END;
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
MATRIX 
t1 100
t2 010
t3 001
;
END;
BEGIN DISTANCES;
DIMENSIONS NTAX=3;
FORMAT TRIANGLE=BOTH MISSING=?;
MATRIX 
t1 0 1.0 2.0
t2 1.0 0 3.0
t3 2.0 3.0 0
;
END;
BEGIN TREES;
TREE 1 = (t1,t2,t3);
END;"""


def test_normalise_rename_taxon2(nexus):
    nex = nexus(
        CHARACTERS="DIMENSIONS NCHAR=3; MATRIX 't 1' 100 t2 010 t3 001;",
        DISTANCES="FORMAT NODIAGONAL; MATRIX 't 1' t2  1.0 t3 2.0 3.0;",
        TREES="TRANSLATE a 't 1', b t2, c t3; TREE 1 = (a,b\n,c);")
    res = str(normalise(nex, rename_taxa=lambda t: t.replace(' ', '_')))
    assert res == """#NEXUS
BEGIN TAXA;
DIMENSIONS NTAX=3;
TAXLABELS t_1 t2 t3;
END;
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
MATRIX 
t_1 100
t2  010
t3  001
;
END;
BEGIN DISTANCES;
DIMENSIONS NTAX=3;
FORMAT TRIANGLE=BOTH MISSING=?;
MATRIX 
t_1 0 1.0 2.0
t2  1.0 0 3.0
t3  2.0 3.0 0
;
END;
BEGIN TREES;
TREE 1 = (t_1,t2,t3);
END;"""


def test_normalise_rename_taxon3(nexus):
    nex = nexus(
        CHARACTERS="DIMENSIONS NCHAR=3; MATRIX 't 1' 100 t2 010 t3 001;",
        DISTANCES="FORMAT NODIAGONAL; MATRIX 't 1' t2  1.0 t3 2.0 3.0;",
        TREES="TRANSLATE a 't 1', b t2, c t3; TREE 1 = (a,b\n,c);")
    res = str(normalise(nex, rename_taxa=lambda t: t.replace(' ', ':')))
    assert res == """#NEXUS
BEGIN TAXA;
DIMENSIONS NTAX=3;
TAXLABELS 't:1' t2 t3;
END;
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
MATRIX 
't:1' 100
t2    010
t3    001
;
END;
BEGIN DISTANCES;
DIMENSIONS NTAX=3;
FORMAT TRIANGLE=BOTH MISSING=?;
MATRIX 
't:1' 0 1.0 2.0
t2    1.0 0 3.0
t3    2.0 3.0 0
;
END;
BEGIN TREES;
TREE 1 = ('t:1',t2,t3);
END;"""


def test_normalise_stripcomments():
    res = normalise(Nexus('#nexus beg[c]in bl[&c]ock; cmd; end[c];'), strip_comments=True)
    assert '[c]' not in str(res)
    assert '[&c]' in str(res)
    assert res.BLOCK





def test_Taxa(nexus):
    nex = nexus(TAXA="""
DIMENSIONS ntax=4;
TAXLABELS
[1] 'John'
[2] 'Paul'
[3] 'George'
[4] 'Ringo'
;""")
    assert nex.TAXA.DIMENSIONS.ntax == 4
    assert list(nex.TAXA.TAXLABELS.labels.values()) == ['John', 'Paul', 'George', 'Ringo']



import io
import shlex
import logging

import pytest

from commonnexus.__main__ import main as cli
from commonnexus import Nexus


@pytest.fixture
def main():
    def f(args):
        return cli(
            shlex.split(args) if isinstance(args, str) else args, log=logging.getLogger(__name__))
    return f


@pytest.fixture
def mainnexus(main, capsys):
    def f(args):
        main(args)
        out = capsys.readouterr()[0]
        return Nexus(out)
    return f


def test_help(main, capsys):
    assert main('') == 1
    out, _ = capsys.readouterr()
    assert 'commonnexus' in out

    with pytest.raises(SystemExit):
        main('help normalise')
    out, _ = capsys.readouterr()
    assert 'normalise' in out

    main('help')


def test_split(main, tmp_path, fixture_dir, caplog):
    with caplog.at_level(logging.INFO):
        main('split --stem test --outdir {} {}'.format(
            tmp_path, fixture_dir / 'multitaxa_mesquite.nex'))
    assert len(caplog.records) == 10
    for rec in caplog.records:
        _, _, fname = rec.message.partition('written to ')
        assert tmp_path.joinpath(fname).exists()


def test_normalise(main, capsys, mocker):
    main('normalise "#nexus begin block; end;"')
    out, _ = capsys.readouterr()
    assert out.strip() == "#NEXUS\nBEGIN block;\nEND;"
    # Make sure reading from stdin works:
    mocker.patch('commonnexus.cli_util.sys.stdin', io.StringIO('#nexus begin block; end;'))
    main('normalise -')
    out, _ = capsys.readouterr()
    assert out.strip() == "#NEXUS\nBEGIN block;\nEND;"


def test_combine(main, capsys, fixture_dir):
    nex = "#nexus begin data; dimensions nchar=1; matrix t1 1 t2 0 t3 1; end;"
    main('combine "{0}{0}" "{0}"'.format(nex))
    out, _ = capsys.readouterr()
    assert '111' in out and '000' in out
    # Combine non-STANDARD datatype from file:
    main('combine {0} {0}'.format(fixture_dir / 'christophchamp_basic.nex'))
    out, _ = capsys.readouterr()
    assert 'MA-LLMA-LL' in out

    main('combine "#nexus begin block; end;" --drop-unsupported')
    out, _ = capsys.readouterr()
    assert out.strip() == '#NEXUS'


def test_taxa(main, mainnexus, capsys, fixture_dir, tmp_path, caplog, morphobank):
    main('taxa {}'.format(fixture_dir / 'christophchamp_basic.nex'))
    out, _ = capsys.readouterr()
    assert len([l for l in out.split('\n') if l.strip()]) == 4

    out = mainnexus('taxa --drop A {}'.format(fixture_dir / 'christophchamp_basic.nex'))
    assert out.taxa == ['B', 'C', 'D']

    out = mainnexus('taxa --rename A,X {}'.format(fixture_dir / 'christophchamp_basic.nex'))
    assert out.taxa == ['X', 'B', 'C', 'D']

    out = mainnexus('taxa --rename "lambda l: chr(ord(l) + 4)" {}'.format(
        fixture_dir / 'christophchamp_basic.nex'))
    assert out.taxa == ['E', 'F', 'G', 'H']

    nex = tmp_path.joinpath('test.nex')
    nex.write_text(str(out), encoding='utf8')
    main('taxa --check {}'.format(str(nex)))
    assert not caplog.records

    nex = tmp_path / 'test.nex'
    nex.write_text("""#nexus
begin taxa; dimensions ntax=4; taxlabels A B C D; end;
begin distances;
matrix A 0 B 1 0 C 2 1 0 D 3 2 1 0;
end;""", encoding='utf8')
    out = mainnexus('taxa --rename A,X {}'.format(nex))
    assert out.taxa == ['X', 'B', 'C', 'D']
    assert 'X' in out.DISTANCES.get_matrix()

    out = mainnexus('taxa --drop A {}'.format(nex))
    assert out.taxa == ['B', 'C', 'D']
    assert list(out.DISTANCES.get_matrix()) == ['B', 'C', 'D']

    nex = tmp_path / 'test.nex'
    nex.write_text("""#nexus
    begin taxa; dimensions ntax=4; taxlabels A B C D; end;
    begin trees; translate 1 A, 2 B, 3 C, 4 D;
    tree 1 = (1,2,3,4);
    end;""", encoding='utf8')
    out = mainnexus('taxa --rename A,X {}'.format(nex))
    assert out.taxa == ['X', 'B', 'C', 'D']
    assert 'X' in out.TREES.TRANSLATE.mapping.values()

    out = mainnexus('taxa --drop A {}'.format(nex))
    assert out.taxa == ['B', 'C', 'D']
    assert not out.TREES.TRANSLATE

    out = mainnexus('taxa --rename A,X "#nexus begin distances; matrix A 0 B 1 0 C 2 1 0; end;"')
    assert 'X' in out.taxa
    out = mainnexus('taxa --drop A "#nexus begin distances; matrix A 0 B 1 0 C 2 1 0; end;"')
    assert list(out.DISTANCES.get_matrix()) == ['B', 'C']

    main('taxa --describe 1 {}'.format(morphobank))
    out, _ = capsys.readouterr()
    assert 'Waldman' in out


def test_taxa_check(main, caplog, tmp_path):
    nex = tmp_path / 'test.nex'
    nex.write_text("""#nexus
    begin taxa; dimensions ntax=3; taxlabels A B C; end;
    begin distances;
    matrix A 0 X 1 0;
    end;""", encoding='utf8')
    with pytest.warns(UserWarning, match='undeclared taxa'):
        main('taxa --check {}'.format(nex))

    nex = tmp_path / 'test.nex'
    nex.write_text("""#nexus
begin taxa; dimensions ntax=4; taxlabels A B C D; end;
begin data;
dimensions nchar=1;
matrix A 1 B 0 C 0 X 0;
end;""", encoding='utf8')
    with pytest.warns(UserWarning, match='undeclared taxa'):
        main('taxa --check {}'.format(nex))
    assert len(caplog.records) == 1
    caplog.clear()

    nex = tmp_path / 'test.nex'
    nex.write_text("""#nexus
begin taxa; dimensions ntax=4; taxlabels A B C D; end;
begin trees;
translate 1 A, 2 B, 3 C, 4 X;
tree 1 = ((Y,A,B)Z);
end;""", encoding='utf8')
    with pytest.warns(UserWarning):
        main('taxa --check {}'.format(nex))
    assert len(caplog.records) == 3


@pytest.mark.parametrize(
    'inchar,matrix,op,args,outchar',
    [
        (1, 't1 a t2 b t3 c', 'binarise', '', 3),
        (2, 't1 10 t2 01 t3 00', 'multistatise', 'k', 1),
        (4, 't1 1001 t2 0111 t3 0010', 'multistatise', '"lambda c: str(int(c)%2)"', 2),
        (2, 't1 10 t2 11 t3 10', 'drop', 'constant', 1),
        (2, 't1 10 t2 11 t3 10', 'drop-numbered', '2', 1),
    ]
)
def test_characters(mainnexus, inchar, matrix, op, args, outchar):
    nex = "#nexus begin data; dimensions nchar={}; matrix {}; end;".format(inchar, matrix)
    res = mainnexus('characters --{} {} "{}"'.format(op, args or '', nex))
    assert res.characters.DIMENSIONS.nchar == outchar


def test_characters_binarise(mainnexus, morphobank):
    """Make sure state labels are used to create charlabels of binarised characters."""
    res = mainnexus('characters --binarise {}'.format(morphobank))
    assert "'Spine of urohyal shape_fused to form single spine'" in str(res)


@pytest.mark.parametrize(
    'inchar,matrix,what,nlines',
    [
        (3, 't1 100 t2 110 t3 101', 'binary-setsize', 3),
        (3, 't1 100 t2 010 t3 001', 'binary-unique', 3),
        (2, 't1 10 t2 11 t3 10', 'binary-constant', 1),
        (3, 't1 100 t2 110 t3 101', 'states-distribution', 5),
    ]
)
def test_characters_describe(main, inchar, matrix, what, nlines, capsys):
    nex = "#nexus begin data; dimensions nchar={}; matrix {}; end;".format(inchar, matrix)
    main('characters --describe {} "{}"'.format(what, nex))
    out, _ = capsys.readouterr()
    lines = [l for l in out.split('\n') if l.strip()]
    assert len(lines) == nlines


def test_characters_convert(main, capsys, fixture_dir):
    cli(['characters', str(fixture_dir / 'christophchamp_basic.nex')])
    main('characters --convert fasta {}'.format(fixture_dir / 'christophchamp_basic.nex'))
    out, _ = capsys.readouterr()
    assert out == """> A
MA-LL
> B
MA-LE
> C
MEATY
> D
ME-TE
"""


@pytest.mark.parametrize(
    'op,args,ntrees',
    [
        ('drop', '1-10', 90),
        ('drop', '1,10', 98),
        ('sample', '10', 10),
        ('random', '5', 5),
    ]
)
def test_trees(op, ntrees, args, mainnexus, tmp_path, nexus):
    trees = tmp_path / 'trees.nex'
    trees.write_text(
        str(nexus(TREES="\n".join('TREE {} = (a,b,c)d;'.format(i + 1) for i in range(100)))))
    nex = mainnexus('trees --{} {} {}'.format(op, args or '', trees))
    assert len(nex.TREES.trees) == ntrees


def test_trees_strip_comments(mainnexus):
    nex = mainnexus('trees --strip-comments "#nexus begin trees; tree 1 = (a[c],b); end;"')
    assert '[c]' not in str(nex)
    assert '(a,b)' in str(nex)


def test_trees_describe(main, capsys):
    main('trees --describe "#nexus begin trees; tree the_tree = (a[c],b); end;"')
    out, _ = capsys.readouterr()
    assert 'the_tree' in out


def test_trees_rename(main, capsys):
    main('trees --rename x,y "#nexus begin trees; tree x = (a[c],b); end;"')
    out, _ = capsys.readouterr()
    assert 'y' in out

    main("""trees --rename "lambda x: '{}1'.format(x[0])" "#nexus begin trees; translate a x, b y, c z; tree xy = (a[c],b); end;" """)
    out, _ = capsys.readouterr()
    assert 'x1' in out



import pytest

from commonnexus import Nexus
from commonnexus.command import Command


@pytest.mark.parametrize(
    'cmd,normalised',
    [
        (' cmd;', '\n\tCMD;'),
        (' [c] cmd;', '\n[c]\n\tCMD;'),
        (' cm[c]d;', '\n\tCMD[c];'),
    ]
)
def test_command(cmd, normalised):
    nex = Nexus('#nexus{}'.format(cmd))
    assert str(Nexus([c.with_normalised_whitespace() for c in nex])) == '#NEXUS' + normalised


def test_serialization():
    cmd = Command.from_name_and_payload('CMD', 'do stuff')
    assert str(cmd) == '\nCMD do stuff;'



import warnings

import pytest

from commonnexus import Nexus, Config
from commonnexus.blocks.characters import GAP, Characters


def test_Eliminate(nexus):
    nex = nexus(CHARACTERS='ELIMINATE 1-3;', config=Config(ignore_unsupported=False))
    with pytest.raises(NotImplementedError):
        _ = nex.CHARACTERS.commands
    _ = nexus(CHARACTERS='ELIMINATE 1-3;').CHARACTERS.ELIMINATE


def test_Chars_1(nexus):
    nex = nexus(
        TAXA="DIMENSIONS  NTAX=3; TAXLABELS A B C;",
        CHARACTERS="""\
DIMENSIONS  NCHAR=3;
FORMAT MISSING=? GAP=- SYMBOLS="AB C" EQUATE="x=(AB)y={BC} z=C" NOLABELS;
CHARSTATELABELS 1  CHAR_A/astate,  2  CHAR_B,  3  CHAR_C/_ bstate cstate;
MATRIX xBC ABC A(A B)C;""")
    assert nex.CHARACTERS.validate()
    assert nex.CHARACTERS.DIMENSIONS.nchar == 3
    assert nex.CHARACTERS.FORMAT.gap == '-'
    assert nex.CHARACTERS.FORMAT.symbols == list('ABC')
    assert nex.CHARACTERS.FORMAT.equate == dict(X=('A', 'B'), Y={'B', 'C'}, Z='C')
    assert nex.CHARACTERS.CHARSTATELABELS.characters[-1].name == 'CHAR_C'
    assert nex.CHARACTERS.CHARSTATELABELS.characters[-1].states == ['_', 'bstate', 'cstate']
    matrix = nex.CHARACTERS.get_matrix(labeled_states=True)
    assert matrix['A']['CHAR_A'] == ('astate', 'B')
    assert matrix['C']['CHAR_C'] == 'cstate'
    assert matrix['C']['CHAR_B'] == ('A', 'B')


@pytest.mark.parametrize(
    'taxa,characters,expect',
    [
        (
            None,
            """DIMENSIONS NCHAR=3; MATRIX t1 001 t2 101 t3 100;""",
            lambda m: list(m['t1'].values()) == ['0', '0', '1'],
        ),
        (
            None,
            "DIMENSIONS NCHAR=3; CHARLABELS c1 c2 c3; MATRIX t1 001 t2 101 t3 100;",
            lambda m: m['t1']['c1'] == '0',
        ),
        (  # With interleaved matrices newlines are important:
            "DIMENSIONS  NTAX=3; TAXLABELS A B C;",
            """DIMENSIONS  NCHAR=3;
FORMAT INTERLEAVE NOLABELS SYMBOLS="ABC";
CHARSTATELABELS 1  CHAR_A,  2  CHAR_B,  3  CHAR_C;
MATRIX
                   AB
                   AB
                   AB
                   C
                   C
                   C
;""",
            lambda m: list(m['A'].values()) == ['A', 'B', 'C'] and m['C']['CHAR_C'] == 'C',
        ),
        (
            None,
            'DIMENSIONS NCHAR=9; FORMAT labels=left SYMBOLS="-+x"; MATRIX t1 (-+){-+}+---+-- t2 +x-++--+x t3 -++++--+x;',
            lambda m: list(m['t1'].values()) == [('-', '+'), {'-', '+'}, '+', '-', '-', '-', '+', '-', '-'],
        ),
        (  # Test MISSING, GAP and MATCHCHAR
            None,
            'DIMENSIONS NCHAR=3; FORMAT MISSING=u GAP=_ MATCHCHAR=.; MATRIX t1 U_0 t2 11. t3 ...;',
            lambda m: [list(r.values()) for r in m.values()] == [[None, GAP, '0'], ['1', '1', '0'], [None, GAP, '0']],
        ),
        (  # Test RESPECTCASE
            None,
            'DIMENSIONS NCHAR=3; FORMAT RESPECTCASE SYMBOLS="ABC" MISSING=c GAP=a MATCHCHAR=b; MATRIX t1 cCa t2 AbB;',
            lambda m: [list(r.values()) for r in m.values()] == [[None, 'C', GAP], ['A', 'C', 'B']],
        ),
        (  # No RESPECTCASE directive, characters are returned as specified in SYMBOLS.
            None,
            'DIMENSIONS NCHAR=3; FORMAT SYMBOLS="ab"; MATRIX t1 aAb t2 BbB;',
            lambda m: [list(r.values()) for r in m.values()] == [['a', 'a', 'b'], ['b', 'b', 'b']],
        ),
        (
            None,
            'DIMENSIONS NCHAR=3; FORMAT DATATYPE=DNA; MATRIX t1 RTA;',
            lambda m: list(m['t1'].values()) == [{'A', 'G'}, 'T', 'A'],
        ),
        (
            None,
            'DIMENSIONS NCHAR=3; FORMAT DATATYPE=NUCLEOTIDE; MATRIX t1 UTA;',
            lambda m: list(m['t1'].values()) == ['T', 'T', 'A'],
        ),
        (
            None,
            'DIMENSIONS NCHAR=3; FORMAT DATATYPE=PROTEIN; MATRIX t1 ABZ;',
            lambda m: list(m['t1'].values()) == ['A', {'D', 'N'}, {'E', 'Q'}],
        ),
        (  # From the documentation of MATCHCHAR:
            None,
            'DIMENSIONS NCHAR = 7; FORMAT DATATYPE=DNA MATCHCHAR = .; MATRIX taxon_1 GACCTTA taxon_2 ...T..C taxon_3 ..T.C..;',
            lambda m: ''.join(m['taxon_2'].values()) == 'GACTTTC',
        ),
        (
            None,
            'dimensions nchar=3; matrix *T 101 OtherT 010;',
            lambda m: '*T' in m,
        ),
        (  # comma-separated polymorphic states.
            None,
           'dimensions nchar=3; matrix t1 (1,0)01 t2 {0,1}10;',
            lambda m: m['t1']['1'] == ('1', '0') and m['t2']['1'] == {'1', '0'},
        ),
    ]
)
def test_Characters_get_matrix(characters, taxa, expect):
    nex = Nexus("""#NEXUS
{}
BEGIN CHARACTERS;
{}
END;""".format('BEGIN TAXA;\n{}\nEND;'.format(taxa) if taxa else '', characters))
    matrix = nex.CHARACTERS.get_matrix()
    assert expect(matrix)


def test_NoDefaultMatchChar():
    nex = Nexus("""#nexus
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT SYMBOLS=".+";
MATRIX t1 ..+ t2 ++. t3 +++;
END;""", no_default_matchchar=True)
    matrix = nex.characters.get_matrix()
    assert matrix['t1'] == {'1': '.', '2': '.', '3': '+'}


def test_Characters_Statelabels():
    nex = Nexus("""#nexus
BEGIN CHARACTERS;
DIMENSIONS NEWTAXA NTAX=3 NCHAR=3;
TAXLABELS t1 t2 t3;
STATELABELS 1 absent present, 2 x y;
MATRIX t1 100 t2 010 t3 001;
END;""")
    matrix = nex.CHARACTERS.get_matrix(labeled_states=True)
    print(matrix)
    assert matrix['t1']['1'] == 'present'
    assert matrix['t1']['2'] == 'x'


def test_Characters_from_data_invalid():
    with pytest.raises(ValueError):  # The MISSING marker is used as value in the matrix.
        _ = Characters.from_data({'t1': {'c1': '?'}})


def test_Characters_from_data(nexus):
    nex = nexus(
        TAXA="DIMENSIONS NTAX=3; TAXLABELS t1 t2 t3;",
        CHARACTERS="""
DIMENSIONS NCHAR=3;
FORMAT TRANSPOSE NOLABELS;
MATRIX {01}00 (01)10 001;""")
    assert nex.characters.is_binary()
    matrix = nex.CHARACTERS.get_matrix()
    nex.replace_block(nex.CHARACTERS, Characters.from_data(matrix))
    assert str(nex) == """#NEXUS
BEGIN TAXA;
DIMENSIONS NTAX=3; TAXLABELS t1 t2 t3;
END;
BEGIN CHARACTERS;
DIMENSIONS NCHAR=3;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
MATRIX 
t1 {01}(01)0
t2 010
t3 001
;
END;"""
    assert nex.characters.get_matrix() == matrix

    nex = Nexus('#nexus')
    matrix = {'t1': {'1': '0', '2': '1'}, 't2': {'1': '1', '2': '0'}}
    nex.append_block(Characters.from_data(matrix, taxlabels=True))
    assert str(nex) == """#NEXUS
BEGIN CHARACTERS;
DIMENSIONS NEWTAXA NTAX=2 NCHAR=2;
FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
TAXLABELS t1 t2;
MATRIX 
t1 01
t2 10
;
END;"""
    assert nex.characters.get_matrix() == matrix

    nex = Nexus('#nexus')
    matrix = {'t1': {'1': 'a', '2': 'B'}, 't2': {'1': 'A', '2': 'b'}}
    nex.append_block(Characters.from_data(matrix, taxlabels=True))
    assert str(nex) == """#NEXUS
BEGIN CHARACTERS;
DIMENSIONS NEWTAXA NTAX=2 NCHAR=2;
FORMAT DATATYPE=STANDARD RESPECTCASE MISSING=? GAP=- SYMBOLS="ABab";
TAXLABELS t1 t2;
MATRIX 
t1 aB
t2 Ab
;
END;"""
    assert nex.characters.get_matrix() == matrix


@pytest.mark.parametrize(
    'commands',
    [
        """
DIMENSIONS NEWTAXA NTAX=3 NCHAR=3;
TAXLABELS t1 t2 t3;
FORMAT TOKENS;
MATRIX t1 1 0 0 t2 0 1 0 t3 0 0 1; 
        """,
        """
DIMENSIONS NEWTAXA NTAX=3 NCHAR=3;
TAXLABELS t1 t2 t3;
FORMAT DATATYPE=CONTINUOUS;
MATRIX t1 100 t2 010 t3 001; 
        """,
        """
DIMENSIONS NEWTAXA NTAX=3 NCHAR=3;
TAXLABELS t1 t2 t3;
FORMAT STATESFORMAT=COUNT;
MATRIX t1 100 t2 010 t3 001; 
        """,
        """
DIMENSIONS NEWTAXA NTAX=3 NCHAR=3;
TAXLABELS t1 t2 t3;
FORMAT ITEMS=SAMPLESIZE;
MATRIX t1 100 t2 010 t3 001; 
        """
    ]
)
def test_Characters_not_implemented(commands, nexus):
    nex = nexus(CHARACTERS=commands, config=Config(ignore_unsupported=False))
    with pytest.raises(NotImplementedError):
        _ = nex.CHARACTERS.commands


def test_Characters_validate(nexus):
    nex = nexus(CHARACTERS="DIMENSIONS NCHAR=1; TAXLABELS t1; MATRIX t1 1;")
    with pytest.raises(ValueError):
        nex.CHARACTERS.validate()


def test_Data_with_Options(nexus):
    nex = nexus(DATA="DIMENSIONS NCHAR=1; OPTIONS GAPMODE=missing; MATRIX t1 1;")
    assert nex.characters.OPTIONS.gapmode == 'missing'


def test_Data_with_duplicate_charlabels(nexus):
    with warnings.catch_warnings(record=True) as w:
        nex = nexus(DATA="DIMENSIONS NCHAR=2; CHARLABELS x x; MATRIX t1 1 1;")
        nex.characters.get_matrix()
        assert len(w) == 1, 'Expected 1 warning, got %r' % w


def test_Data_with_mixed_charlabels(nexus):
    nex = nexus(DATA="DIMENSIONS NCHAR=2; CHARSTATELABELS 1 x, 2 ; MATRIX t1 1 1;")
    m = nex.characters.get_matrix()
    assert '2' in m['t1'], 'unspecified character label'


def test_illegal_charstatelabel():
    nex = Nexus("""\
#NEXUS

BEGIN DATA;
    DIMENSIONS NTAX=3 NCHAR=3;
    FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
    CHARSTATELABELS
        1 hand_1,
        2 burn(tr.)_3,
        3 claw(nail)_3
        ;
MATRIX
A   001
B   000
C   000
;
END;""")
    with pytest.raises(ValueError) as e:
        _ = nex.DATA.CHARSTATELABELS
    assert 'burn(' in str(e)

    with pytest.raises(ValueError):
        nex.validate()


def test_duplicate_charstatelabels():
    nex = Nexus("""\
#NEXUS

BEGIN DATA;
    DIMENSIONS NTAX=3 NCHAR=3;
    FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
    CHARSTATELABELS
        1 hand_1,
        2 _3,
        3 _3
        ;
MATRIX
A   001
B   000
C   000
;
END;""")
    with warnings.catch_warnings(record=True) as w:
        _ = nex.DATA.CHARSTATELABELS
        assert len(w) == 1, 'Expected 1 warning, got %r' % w

    nex.cfg.strict = True
    with pytest.raises(ValueError):
        _ = nex.DATA.CHARSTATELABELS



"""Regression Tests"""
import warnings

import pytest

from commonnexus import Nexus
from commonnexus.blocks import Characters


@pytest.fixture
def regression(fixture_dir):
    return fixture_dir / 'regression'


def test_UnquotedSymbols(regression):
    with warnings.catch_warnings(record=True) as w:
        nex = Nexus.from_file(regression / 'unquoted_symbols.nex')
        assert nex.DATA.FORMAT.symbols == ['0', '1']
        assert len(w) == 72, 'Expected 72 warnings, got %r' % w


def test_Morphobank(morphobank, regression):
    nex = Nexus.from_file(morphobank)
    assert len(nex.NOTES.texts) == 30
    notes = nex.NOTES.get_texts(character='Lateral ethmoid-lacrimal articulation, orientation')
    assert len(notes) == 1
    assert notes[0].text == 'Waldman, 1986'

    matrix = nex.characters.get_matrix(labeled_states=True)
    assert matrix['Sargocentron vexillarium']['Vomer, shape of tooth patch'] == 'Trapezoidal to ovate'

    m = nex.characters.get_matrix()
    nex = Nexus.from_blocks(
        Characters.from_data(m, statelabels=nex.characters.get_charstatelabels()[1]))
    assert nex.characters.get_matrix(labeled_states=True) == matrix

    nex = Nexus.from_file(regression / 'mbank_X29171_3-10-2023_150.nex')
    m = nex.characters.get_matrix(labeled_states=True)
    assert m['Proterosuchus fergusi']['Skull and lower jaws, interdental plates'] == \
           ('absent', 'present, small and well-spaced from each other')


def test_treebase(regression):
    nex = Nexus.from_file(regression / 'treebase-matrix-123.nex')
    assert len(nex.characters.get_matrix()) == 64
    nex = Nexus.from_file(regression / 'treebase-tree-1234.nex')
    assert set(nex.taxa ) == {n.name for n in nex.TREES.TREE.newick.walk() if n.is_leaf}


def test_WhitespaceInMatrix_regression(regression):
    """Regression: Test that leading whitespace in a data matrix is parsed ok"""
    nex = Nexus.from_file(regression / 'white_space_in_matrix.nex')
    matrix = nex.DATA.get_matrix()
    assert len(list(matrix.values())[0]) == 2
    assert list(matrix['Harry'].values()) == ['0', '0']
    assert list(matrix['Simon'].values()) == ['0', '1']
    assert list(matrix['Betty'].values()) == ['1', '0']
    assert list(matrix['Louise'].values()) == ['1', '1']


def test_RandomAPETrees_regression(regression):
    """Regression: Test that we can parse randomly generated APE/R trees"""
    nex = Nexus.from_file(regression / 'ape_random.trees')
    assert len(nex.TREES.trees) == 2
    assert nex.TREES.TREE.name == 'UNTITLED'


def test_BadCharsInTaxaName_trees(regression):
    bad_chars = Nexus.from_file(regression / 'bad_chars_in_taxaname.trees')
    assert bad_chars.TREES, 'No TREES block found'
    assert len(bad_chars.TREES.trees) == 1, 'Invalid number of TREEs found'

    # did we get the translation parsed properly.
    translated = bad_chars.TREES.translate(bad_chars.TREES.TREE)
    assert 'PALAUNGWA_De.Ang' in {n.name for n in translated.walk()}
    expected = '(MANGIC_Bugan,MANGIC_Paliu,MANGIC_Mang,PALAUNGWA_Danaw,PAL'
    assert expected in translated.newick


def test_DetranslateWithDash_trees(regression):
    nex = Nexus.from_file(regression / 'detranslate-with-dash.trees', hyphenminus_is_text=True)
    translated = nex.TREES.translate(nex.TREES.TREE)
    assert "(one,two,three,four-1,four_2)" in translated.newick


def test_BranchLengthsInIntegers_trees(regression):
    int_length = Nexus.from_file(regression / 'branchlengths-in-integers.trees')
    translated = int_length.TREES.translate(int_length.TREES.TREE)
    assert '(one:0.1,two:0.2,three:1,four:3,five:0.3)' in translated.newick


def test_Mesquite_taxa_block(regression):
    """Regression: Test that we can parse MESQUITE taxa blocks"""
    mesquite = Nexus.from_file(regression / 'mesquite_taxa_block.nex')
    assert all(taxon in mesquite.TAXA.TAXLABELS.labels.values() for taxon in ['A', 'B', 'C'])
    assert 'Untitled_Block_of_Taxa'.upper() == mesquite.TAXA.title
    assert 'Untitled_Block_of_Taxa'.upper() == mesquite.TAXA.links['TAXA']


def test_TreeHandler_Regression_Mesquite_attributes(regression):
    mesquite_branches = Nexus.from_file(regression / 'mesquite_formatted_branches.trees')
    assert mesquite_branches.TREES.title == 'Trees from "temp.trees"'.upper()
    assert mesquite_branches.TREES.links['TAXA'] == "Untitled_Block_of_Taxa".upper()
    assert mesquite_branches.TREES.TREE
    translated = mesquite_branches.TREES.translate(mesquite_branches.TREES.TREE)
    assert {'A', 'B', 'C'} == {n.name for n in translated.walk() if n.name}


def test_TaxaHandler_OneLine_block_find(regression):
    """Regression: Test that we can read taxalabels delimited by spaces not new lines."""
    taxlabels = Nexus.from_file(regression / 'taxlabels.nex')
    assert taxlabels.TAXA
    assert taxlabels.CHARACTERS
    matrix = taxlabels.CHARACTERS.get_matrix()
    for taxon in ['A', 'B', 'C']:
        assert taxon in matrix
        assert taxon in taxlabels.TAXA.TAXLABELS.labels.values()

    assert list(matrix["A"].values()) == ["1", "1", "1"]
    assert list(matrix["B"].values()) == ["0", "1", "1"]
    assert list(matrix["C"].values()) == ["0", "0", "1"]


def test_GlottologTree_block_find(regression):
    """Regression: Test that we can read glottolog tree format."""
    glottolog = Nexus.from_file(regression / 'glottolog.trees')
    assert glottolog.TAXA
    assert glottolog.TREES
    for i in range(2, 6):
        taxon = 'abun125%d' % i
        assert taxon in glottolog.TAXA.TAXLABELS.labels.values()

    assert len(glottolog.TAXA.TAXLABELS.labels) == 4
    assert glottolog.TREES.TREE


def test_TreeHandler_MrBayes(regression):
    """
    Test reading of treefile generated by a 2003-era MrBayes which, for some reason, have
    a tab between "tree\t<name>=(...)"
    """
    nex = Nexus.from_file(regression / 'mrbayes.trees')
    assert nex.TREES.TREE.name == 'tree.2500000.34524.300436'


def test_DataHandler_trailing_comma():
    """Trailing comma in charstatelabels creates an empty character at N+1"""
    nex = Nexus("""#NEXUS
        BEGIN DATA;
            DIMENSIONS NTAX=3 NCHAR=1;
            FORMAT MISSING=? GAP=- SYMBOLS="01";
            CHARSTATELABELS
            1 ALL,
        ;

        MATRIX
        A        1
        B        0
        C        0
        ;
        END;
        """)
    with warnings.catch_warnings(record=True) as w:
        assert nex.DATA.CHARSTATELABELS
        assert len(w) == 1, 'Expected 1 warning, got %r' % w
        assert issubclass(w[0].category, UserWarning)
        assert len(nex.DATA.CHARSTATELABELS.characters) == 1

    
def test_TreeHandler_Taxon_with_asterisk(regression):
    """
    Test reading of treefile that contains a taxon with an asterisk in it.
    """
    nex = Nexus.from_file(regression / 'tree_with_asterisk_in_taxa.trees', asterisk_is_text=True)
    assert nex.TREES.TREE
    assert len(nex.TAXA.TAXLABELS.labels) == 38
    assert "*R35" in nex.TAXA.TAXLABELS.labels.values()
    translated = nex.TREES.translate(nex.TREES.TREE)
    assert "*R35" in {n.name for n in translated.walk()}


def test_TreeHandler_TranslateBlockMismatch(regression):
    """
    Test that a warning is generated when a tree has un-translatable nodes.
    """
    nex = Nexus.from_file(regression / 'tree_translate_mismatch.trees')
    with pytest.warns(UserWarning):
        _ = nex.TREES.translate(nex.TREES.TREE)


def test_TreeHandler_TranslateBlockWithComments(regression):
    """
    Test reading of treefile that contains no translate block, so taxa are 
    identified from parsing the first tree. If the tree contains comments of the
    form "x(y)" then these used to get parsed incorrectly (as extra taxa).
    """
    nex = Nexus.from_file(regression / 'tree_default_translate_with_comments.trees')
    assert nex.TREES.TREE
    taxa = set(nex.TAXA.TAXLABELS.labels.values())
    for node in nex.TREES.TREE.newick.walk():
        if node.name:
            assert node.name in taxa


def test_TreeHandler_BeastTranslate(regression):
    """
    Test handling of BEAST Translate Block
    """
    nex = Nexus.from_file(regression / 'detranslate-beast.trees')
    assert nex.TREES.TREE
    assert len(nex.TREES.TRANSLATE.mapping) == 19
    translated = nex.TREES.translate(nex.TREES.TREE)
    leafs = {n.name for n in translated.walk() if n.is_leaf}
    assert len(leafs) == 19
    assert "B" in leafs


def test_TreeHandler_BeastTranslate_2(regression):
    """
    Test handling of BEAST Translate Block in Phlorest.
    """
    nex = Nexus.from_file(regression / 'detranslate-beast-2.trees')
    assert nex.TREES.TREE
    assert len(nex.TREES.TRANSLATE.mapping) == 10
    translated = nex.TREES.translate(nex.TREES.TREE)
    leafs = {n.name for n in translated.walk() if n.is_leaf}
    assert len(leafs) == 10
    assert "B" in leafs


def test_TreeHandler_BeastTranslate_3(regression):
    """
    Test handling of BEAST Translate Block in Phlorest. Again.
    """
    nex = Nexus.from_file(regression / 'tree_detranslate_chang.trees')
    assert nex.TREES.TREE
    translated = nex.TREES.translate(nex.TREES.TREE)
    leafs = {n.name for n in translated.walk() if n.is_leaf}
    assert len(leafs) == 52
    assert "Old_Church_Slavic" in leafs


def test_Edictor(regression):
    """
    Test handling of EDICTOR Nexus files
    """
    nex = Nexus.from_file(regression / 'edictor.nex')
    assert nex.DATA
    assert nex.CHARACTERS
    
    assert nex.DATA.DIMENSIONS.nchar == 4
    assert nex.DATA.DIMENSIONS.ntax == 3

    #assert nex.blocks['characters'].nchar == 4
    #assert nex.blocks['characters'].ntaxa == 0
    #assert repr(nex.blocks['characters']), 'repr of characters block failed'
    
    #assert len(nex.data.charlabels) == 4
    #assert nex.data.nchar == 4
    
    #assert nex.data.charlabels == nex.characters.charlabels



import pytest

from commonnexus.tools.matrix import CharacterMatrix
from commonnexus.blocks.characters import GAP


def test_multistatise(nexus):
    nex = nexus(DATA="""
        Dimensions ntax=4 nchar=4;
        Format datatype=standard symbols="01" gap=-;
        Matrix
        Harry              0000
        Simon              0100
        Betty              0010
        Louise             1001
        ;""")
    res = CharacterMatrix.multistatised(nex.DATA.get_matrix())
    res = {k: v['1'] for k, v in res.items()}
    assert res == dict(Harry=None, Simon='B', Betty='C', Louise=('A', 'D'))


def test_multistatise_with_all_zero_taxa(nexus):
    """Include taxa that have no entries"""
    nex = nexus(DATA="""
        DIMENSIONS  NTAX=15 NCHAR=7;
        FORMAT DATATYPE=STANDARD MISSING=? GAP=- INTERLEAVE=YES;
    MATRIX

    Gertrude                0000001
    Debbie                  0001000
    Zarathrustra            0000000
    Christie                0010000
    Benny                   0100000
    Bertha                  0100000
    Craig                   0010000
    Fannie-May              0000010
    Charles                 0010000
    Annik                   1000000
    Frank                   0000010
    Amber                   1000000
    Andreea                 1000000
    Edward                  0000100
    Donald                  0001000
    ;""")
    m = CharacterMatrix(nex.DATA.get_matrix())
    assert len(list(m.iter_rows())) == 15
    assert 'AAAAAAB' in m.to_fasta()
    assert CharacterMatrix.from_fasta(m.to_fasta())['Amber']['1'] == 'B'
    msnex = CharacterMatrix.multistatised(m)

    for taxon, sites in msnex.items():
        if taxon[0] == 'Z':
            continue  # will check later

        # first letter of taxa name is the expected character state
        assert taxon[0] == sites['1'], "%s should be %s not %s" % (taxon, taxon[0], sites[0])
    # deal with completely missing taxa
    assert 'Zarathrustra' in msnex
    assert msnex['Zarathrustra']['1'] is None
    assert '> Donald' in msnex.to_fasta()
    assert 'Donald    ' in msnex.to_phylip()


def test_CharacterMatrix(nexus):
    nex = nexus(DATA="""
DIMENSIONS NCHAR=5;
FORMAT DATATYPE=STANDARD MISSING=x GAP=y;
MATRIX
t1 1y01(01)
t2 10{01}yx;
""")
    matrix = CharacterMatrix(nex.characters.get_matrix())
    assert matrix.has_missing
    assert matrix.has_gaps
    assert matrix.has_polymorphic
    assert matrix.has_uncertain
    m = CharacterMatrix.from_characters(matrix, drop_uncertain=True, drop_polymorphic=True)
    assert len(m.characters) == 3
    assert m.to_fasta()
    m = CharacterMatrix.from_characters(matrix, drop_constant=True)
    assert len(m.characters) == 4

    with pytest.raises(Exception):
        matrix.to_fasta()

    with pytest.raises(Exception):
        matrix.to_phylip()


@pytest.mark.parametrize(
    'row,kw',
    [
        (['0', '1', None], dict(drop_missing=True)),
        (['0', GAP, '1'], dict(drop_gapped=True)),
        (['0', '1'], dict(drop_chars=['c'])),
        (['0', '1'], dict(drop_chars=['x'], inverse=True)),
    ]
)
def test_from_characters(row, kw):
    m = CharacterMatrix.from_characters(
        {'t{}'.format(i + 1): dict(c=v) for i, v in enumerate(row)}, **kw)
    assert not m.characters


def test_CharacterMatrix_from_fasta():
    with pytest.raises(AssertionError):
        CharacterMatrix.from_fasta('> t1\nABB\n> t2\nAB')


def test_CharMatrix_binarise(nexus):
    matrix = CharacterMatrix.binarised(nexus(
        TAXA="DIMENSIONS NTAX=3; TAXLABELS Maori Dutch Latin;",
        CHARACTERS="""
    Dimensions ntax=3 nchar=3;
    Format datatype=standard symbols="1234567" gap=- transpose nolabels equate="x=(12)";
    Charstatelabels
        1 char1, 2 char2, 3 char3;
    Matrix x23 4-5 67? ;""").characters.get_matrix())
    assert matrix['Maori'] == {
        'char1_1': '1', 'char1_2': '1', 'char1_3': '0',
        'char2_4': '1', 'char2_5': '0',
        'char3_6': '1', 'char3_7': '0',
    }


@pytest.mark.parametrize(
    'matrix,rows',
    [
        ('t1 A t2 B t3 C', (['1', '0', '0'], ['0', '1', '0'], ['0', '0', '1'])),
        ('t1 (AB) t2 B', (['1', '1'], ['0', '1'])),
        ('t1 {AB} t2 B', ([{'1'}, {'1'}], ['0', '1'])),
    ]
)
def test_binarise(nexus, matrix, rows):
    nex = nexus(
        CHARACTERS='DIMENSIONS nchar=1; FORMAT datatype=standard symbols="ABC"; MATRIX {};'.format(
            matrix))
    bin = CharacterMatrix.binarised(nex.characters.get_matrix())
    for row, expected in zip(bin.iter_rows(), rows):
        assert row == expected



import pytest

from commonnexus import Nexus
from commonnexus.blocks import Block


@pytest.mark.parametrize(
    'cmds,kw,expect',
    [
        ([('cmd', "'long name'"), 'cmd2', ('cmd3', '', 'comment')], {},
         lambda b, n: len(b.CMD._tokens) == 1 and b.CMD2 and '[comment]' in n),
        ([], dict(TITLE="the title", ID='012def'),
         lambda b, n: b.title == 'the title'.upper() and b.id == '012def'),
        ([], dict(LINK="taxa = above"),
         lambda b, n: 'TAXA' in b.links and 'LINK taxa = above;' in n),
        ([], dict(LINK=("taxa", "above")),
         lambda b, n: 'TAXA' in b.links),
    ]
)
def test_Block_with_nexus(cmds, kw, expect):
    nex = Nexus()
    nex.append_block(Block.from_commands(cmds, nexus=nex, **kw))
    assert expect(nex.BLOCK, str(nex))


def test_Block_as_string():
    assert str(Block.from_commands([])) == '\nBEGIN BLOCK;\nEND;'



import decimal

import pytest

from commonnexus import Nexus
from commonnexus.blocks.distances import Distances


@pytest.mark.parametrize(
    'matrix',
    [
        """FORMAT TRIANGLE=UPpER NOlABELS;
            MATRIX
                0.0  1.0  2.0  4.0  7.0
                     0.0  3.0  5.0  8.0
                          0.0  6.0  9.0
                               0.0 10.0
                                    0.0;""",
        """FORMAT TRIANGLE=UPpER NOlABELS INTERLEAVE;
            MATRIX
                0.0  1.0  2.0
                0.0  3.0  5.0
                0.0  6.0  9.0
                0.0 10.0
                0.0
                4.0  7.0
                8.0;""",
        """FORMAT TRIANGLE=UPPER;
            MATRIX
                taxon_1 0.0  1.0  2.0  4.0  7.0
                taxon_2      0.0  3.0  5.0  8.0
                taxon_3           0.0  6.0  9.0
                taxon_4                0.0 10.0
                taxon_5                     0.0;""",
        """FORMAT NODIAGONAL NOLabels TRIANGLE=UPPER;
            MATRIX
                1.0  2.0  4.0  7.0
                     3.0  5.0  8.0
                          6.0  9.0
                              10.0;""",
        """FORMAT INTERLEAVE;
            MATRIX
                taxon_1  0.0
                taxon_2  1.0  0.0
                taxon_3  2.0  3.0  0.0
                taxon_4  4.0  5.0  6.0
                taxon_5  7.0  8.0  9.0
                taxon_4  0.0
                taxon_5 10.0  0.0;""",
        """FORMAT NOLABELS INTERLEAVE;
            MATRIX
                0.0
                1.0  0.0
                2.0  3.0  0.0
                4.0  5.0  6.0
                7.0  8.0  9.0
                0.0
                10.0  0.0;""",
        """FORMAT NOLABELS NODIAGONAL INTERLEAVE;
            MATRIX
                1.0
                2.0  3.0
                4.0  5.0  6.0
                7.0  8.0  9.0
                10.0;""",
        """MATRIX
                taxon_1  0.0
                taxon_2  1.0  0.0
                taxon_3  2.0  3.0  0.0
                taxon_4  4.0  5.0  6.0  0.0
                taxon_5  7.0  8.0  9.0 10.0  0.0;""",
        """FORMAT NODIAGONAL;
            MATRIX
                taxon_1
                taxon_2  1.0
                taxon_3  2.0  3.0
                taxon_4  4.0  5.0  6.0
                taxon_5  7.0  8.0  9.0 10.0;""",
        """FORMAT NODIAGONAL NOLABELS;
            MATRIX
                1.0
                2.0  3.0
                4.0  5.0  6.0
                7.0  8.0  9.0 10.0;""",
        """FORMAT NOdiagonal TRIANGLE=BOTH;
            MATRIX
                taxon_1    1.0  2.0  4.0  7.0
                taxon_2  1.0    3.0  5.0  8.0
                taxon_3  2.0  3.0    6.0  9.0
                taxon_4  4.0  5.0  6.0   10.0
                taxon_5  7.0  8.0  9.0 10.0;"""
        """FORMAT TRIANGLE=BOTH;
            MATRIX
                taxon_1  0    1.0  2.0  4.0  7.0
                taxon_2  1.0  0    3.0  5.0  8.0
                taxon_3  2.0  3.0  0    6.0  9.0
                taxon_4  4.0  5.0  6.0  0   10.0
                taxon_5  7.0  8.0  9.0 10.0  0;"""
    ]
)
def test_Distances_get_matrix(matrix):
    for fmt in [
        """\
BEGIN TAXA;
    DIMENSIONS NTAX=5;
    TAXLABELS taxon_1 taxon_2 taxon_3 taxon_4 taxon_5;
END;
BEGIN DISTANCES;
    {}
END;""",
        """\
BEGIN DISTANCES;
    DIMENSIONS NEWTAXA NTAX=5;
    TAXLABELS taxon_1 taxon_2 taxon_3 taxon_4 taxon_5;
    {}
END;"""]:
        dist = Nexus("#NEXUS\n{}".format(fmt.format(matrix))).DISTANCES
        m = dist.get_matrix()
        assert m['taxon_4'] == dict(
            taxon_1=decimal.Decimal('4.0'),
            taxon_2=decimal.Decimal('5.0'),
            taxon_3=decimal.Decimal('6.0'),
            taxon_4=decimal.Decimal('0.0'),
            taxon_5=decimal.Decimal('10.0'),
        )
        nex = Nexus()
        nex.append_block(Distances.from_data(m, nexus=nex))
        assert m == nex.DISTANCES.get_matrix()


def test_minimal():
    dist = Nexus('#nexus begin distances; matrix t1 0 t2 1 0 t3 1 2 0; end;').DISTANCES
    assert [int(e) for e in dist.get_matrix()['t1'].values()] == [0, 1, 1]


def test_missing():
    nex = Nexus('#nexus begin distances; format missing=x; matrix t1 0 t2 x 0 t3 1 2 0; end;')
    assert [None if e is None else int(e)
            for e in nex.DISTANCES.get_matrix()['t1'].values()] == [0, None, 1]


def test_incomplete(nexus):
    nex = nexus(
        TAXA="DIMENSIONS ntax=4; taxlabels A B C D;",
        DISTANCES='matrix A 0 C 1 0 D 2 1 0;')
    matrix = nex.DISTANCES.get_matrix()
    assert 'B' not in matrix
    assert matrix['D'] == dict(A=2, C=1, D=0)


def test_splitstree(fixture_dir):
    nex = Nexus.from_file(fixture_dir / 'woodmouse.nxs')
    assert nex.DISTANCES
    assert nex.DISTANCES.get_matrix()['No305']['No1208S'] == decimal.Decimal('0.018828452')


def test_Distances_from_data():
    nex = Nexus()
    nex.append_block(Distances.from_data({'t1': {'t1': 0}}, taxlabels=True))
    assert str(nex) == """#NEXUS
BEGIN DISTANCES;
DIMENSIONS NEWTAXA NTAX=1 NTAX=1;
FORMAT TRIANGLE=BOTH MISSING=?;
TAXLABELS t1;
MATRIX 
t1 0
;
END;"""

    nex = Nexus()
    nex.append_block(Distances.from_data({'t 1': {'t 1': 0}}, taxlabels=True, nexus=nex))
    assert str(nex) == """#NEXUS
BEGIN DISTANCES;
DIMENSIONS NEWTAXA NTAX=1 NTAX=1;
FORMAT TRIANGLE=BOTH MISSING=?;
TAXLABELS 't 1';
MATRIX 
't 1' 0
;
END;"""
